# SmallOverlap GAP Package
# Copyright (C) 2007-2008 Mark Kambites, University of Manchester
#
# This package contains code for manipulating semigroups and monoids with
# presentations satisfying small overlap conditions of the kind
# introduced by Remmers.  The algorithms used are described in the paper
# M. Kambites, "Small overlap monoids: the word problem", arXiv ref.
#
# This file contains general routines for small overlap semigroups and
# monoids
#############################################################################
# The attributes MaxPiecePrefixLengths, MiddleWordLengths,
# MaxPieceSuffixLengths are lists of the lengths of the maximum piece
# prefixes, middle words, and maximum piece suffixes of the relation words
# in a presentation. See paper for definitions of these terms.
 
InstallMethod(MaxPiecePrefixLengths,
              "for finitely presented monoids and semigroups",
              [IsFpMonoidOrSemigroup], 
   function(S)
       local relwords;
       relwords := RelationWords(S);
       return List(relwords,
                   x -> LengthLongestPieceFactorToRight(relwords, x, 1));
   end
);

InstallMethod(MaxPieceSuffixLengths,
              "for finitely presented monoids and semigroups",
              [IsFpMonoidOrSemigroup], 
   function(S)
       local relwords;
       relwords := RelationWords(S);
       return List(relwords,
                   x -> LengthLongestPieceFactorToLeft(relwords,x,Length(x)));
   end
);

InstallMethod(MiddleWordLengths,
              "for finitely presented monoids and semigroups",
              [IsFpMonoidOrSemigroup], 
   function(S)
       return RelationWordLengths(S) - MaxPiecePrefixLengths(S)
                                             - MaxPieceSuffixLengths(S);
   end
);

#############################################################################
# The attribute MaxPieceSuffixLengthBound is the length of the longest maximal
# piece suffix of any relation word in the presentation. It is useful for
# creating a buffer large enough for the computations involved in solving
# the word problem.

InstallMethod(MaxPieceSuffixLengthBound,
              "for finitely presented monoids and semigroups",
              [IsFpMonoidOrSemigroup], 
   function(S)
       local m;
       m := MaxPieceSuffixLengths(S);
       if IsEmpty(m) then
           return 0;
       else
           return Maximum(m);
       fi;
   end
);

############################################################################
# Attribute SmallOverlapClass is the small overlap class of an fp monoid
# or semigroup S, that is, the greatest integer positive n such that the
# defining presentation for S satisfies C(n), or infinity if C(n) is
# satisfied for all n, or 0 if C(n) is not satisfied for any n.

InstallMethod(SmallOverlapClass,
              "for finitely presented monoids and semigroups",
              [IsFpMonoidOrSemigroup],
              function(S)
    local relwords, l, r, n, p, f; 

    relwords := RelationWords(S);

    # We attempt to factorise each relation in turn as a product of pieces
    # of maximal length, starting from the left.
    l := infinity; # Smallest number of pieces to cover a relation word so far
    for r in relwords do
        p := 1; # Working position in the word r
        n := 0; # Number of pieces required to cover r so far
        while n <> infinity and p <= Length(r) do
            f := LengthLongestPieceFactorToRight(relwords, r, p);
            if f = 0 then
                # This happens if the letter relwords[p] is not a piece, that
                # is, occurs only once across all the relation words, in which
                # case this relation word is not a product of pieces at all.
                n := infinity;
            else
                n := n + 1;
                p := p + f;
            fi;
        od;

        # We are interested in the smallest number of pieces required to
        # cover ANY relation word, that is, the minimum value of n over
        # all relation words r.
        l := Minimum(n, l);
     od;

     # While we are at it, we explicitly set the properties IsC4,
     # IsC3, IsC2 and IsC1.
     SetIsC4(S, l >= 4);
     SetIsC3(S, l >= 3);
     SetIsC2(S, l >= 2);
     SetIsC1(S, l >= 1);
     return l;
end);


##########################################################################
# Properties IsC4(S), IsC3(S), IsC2(S), IsC1(S) are (logically) shorthand
# for SmallOverlapClass(S) >= 4, 3, 2, 1 respectively. They are currently
# just implemented using calls to SmallOverlapClass. They could be made
# more efficient by using dedicated code, and this might be worth doing
# if you are planning to work with enormous presentations, but for
# "everyday"-sized presentations, SmallOverlapClass is so fast that it
# doesn't seem worth the effort.

InstallMethod(IsC4, "for finitely presented monoids and semigroups", [IsFpMonoidOrSemigroup], 
   function(S)
   return SmallOverlapClass(S) >= 4;
end);

InstallMethod(IsC3, "for finitely presented monoids and semigroups", [IsFpMonoidOrSemigroup], 
   function(S)
   return SmallOverlapClass(S) >= 3;
end);

InstallMethod(IsC2, "for finitely presented monoids and semigroups", [IsFpMonoidOrSemigroup], 
   function(S)
   return SmallOverlapClass(S) >= 2;
end);

InstallMethod(IsC1, "for finitely presented monoids and semigroups", [IsFpMonoidOrSemigroup], 
   function(S)
   return SmallOverlapClass(S) >= 1;
end);

# We may as well tell GAP that C(4) => C(3) => C(2) => C(1)
InstallTrueMethod(IsC1, IsC2);
InstallTrueMethod(IsC2, IsC3);
InstallTrueMethod(IsC3, IsC4);

##########################################################################
# IsElementOfC4, IsElementOfC3, IsElementOfC2, IsElementOfC1 are methods
# to test, for an given element of a finitely presented monoid or semigroup,
# whether the containing monoid or semigroup satisfies C(4), C(3), C(2), C(1)
# respectively.

InstallMethod(IsElementOfC4, "for everything",
              [IsElementOfFpMonoidOrSemigroup], function(u)
    return IsC4(FpGrpMonSmgOfFpGrpMonSmgElement(u));
end);

InstallMethod(IsElementOfC3, "for everything",
              [IsElementOfFpMonoidOrSemigroup], function(u)
    return IsC3(FpGrpMonSmgOfFpGrpMonSmgElement(u));
end);

InstallMethod(IsElementOfC2, "for everything", 
              [IsElementOfFpMonoidOrSemigroup], function(u)
    return IsC2(FpGrpMonSmgOfFpGrpMonSmgElement(u));
end);

InstallMethod(IsElementOfC1, "for everything",
              [IsElementOfFpMonoidOrSemigroup], function(u)
    return IsC1(FpGrpMonSmgOfFpGrpMonSmgElement(u));
end);

# Again, since C4 => C3 => C2 => C1 we have....
InstallTrueMethod(IsElementOfC1, IsElementOfC2);
InstallTrueMethod(IsElementOfC2, IsElementOfC3);
InstallTrueMethod(IsElementOfC3, IsElementOfC4);


############################################################################
# LengthLongestPieceFactorToRight takes a list of relation words, another
# word w and a position. It returns the length of the longest factor of w
# starting at position p which is a piece (ie. which occurs twice in the
# relation words). This may of course be 0!

InstallGlobalFunction(LengthLongestPieceFactorToRight,
                      function(relwords, w, p)
    local maxonce, maxtwice, r, i, n;
    maxonce := 0;   # Length of longest factor we have found in
                    # the relation words so far
    maxtwice := 0;  # Length of longest factor we have found TWICE in
                    # the relation words so far
    for r in relwords do
        for i in [1..Length(r)] do
            n := 0;
            while p+n <= Length(w) and i+n <= Length(r) and w[p+n] = r[i+n] do
               n := n + 1;
            od;
            maxtwice := Maximum(Minimum(n, maxonce), maxtwice);
            maxonce := Maximum(n, maxonce);
         od;
     od;
     return maxtwice;
end);

############################################################################
# LengthLongestPieceFactorToLeft is much as LengthLongestPieceFactorToRight,
# but computes the length of the longest piece factor of w ENDING at position
# p.

InstallGlobalFunction(LengthLongestPieceFactorToLeft, function(relwords, w, p)
    local maxonce, maxtwice, r, i, n;
    maxonce := 0;
    maxtwice := 0;
    for r in relwords do
        for i in [1..Length(r)] do
            n := 0;
            while p-n >= 1 and i-n >= 1 and w[p-n] = r[i-n] do
               n := n + 1;
            od;
            maxtwice := Maximum(Minimum(n, maxonce), maxtwice);
            maxonce := Maximum(n, maxonce);
         od;
     od;
     return maxtwice;
end);

#############################################################################
# Install ViewObj methods for small overlap monoids and semigroups

InstallMethod(ViewObj,
              "for small overlap finitely presented monoids",
              [IsFpMonoid and HasSmallOverlapClass], 1200, function(S)
    if SmallOverlapClass(S) >= 1 then
        Print("<C(", SmallOverlapClass(S),
              ") small overlap fp monoid on the generators ",
              FreeGeneratorsOfFpMonoid(S),">");
    else
        TryNextMethod();
    fi;
end);

InstallMethod(ViewObj,
              "for small overlap finitely presented semigroups",
              [IsFpSemigroup and HasSmallOverlapClass], 1200, function(S)
    if SmallOverlapClass(S) >= 1 then
        Print("<C(", SmallOverlapClass(S),
              ") small overlap fp monoid on the generators ",
              FreeGeneratorsOfFpMonoid(S),">");
    else
        TryNextMethod();
    fi;
end);

############################################################################
# The following are explicit ViewObj methods for C4, C3, C2 and C1
# monoids and semigroups. At present they are not likely to be used, since
# the HasSmallOverlapClass method takes precedence, and GAP never discovers
# that a semigroup has one of these properties without finding out its exact
# small overlap class. However, they will become useful if more efficient
# means are used to test and set IsC4 etc without necessarily calculating the
# exact small overlap class.

InstallMethod(ViewObj, "for C4 finitely presented monoids",
             [IsC4 and IsFpMonoid], 1000, function(S)
       Print("<at least C(4) small overlap fp monoid on the generators ",
             FreeGeneratorsOfFpMonoid(S),">");
end);

InstallMethod(ViewObj, "for C4 finitely presented semigroups",
              [IsC4 and IsFpSemigroup], 1000, function(S)
       Print("<at least C(4) small overlap fp semigroup on the generators ",
             FreeGeneratorsOfFpSemigroup(S),">");
end);

InstallMethod(ViewObj, "for C3 finitely presented monoids",
              [IsC3 and IsFpMonoid], 990, function(S)
       Print("<at least C(3) small overlap fp monoid on the generators ",
             FreeGeneratorsOfFpMonoid(S),">");
end);

InstallMethod(ViewObj, "for C3 finitely presented semigroups",
              [IsC3 and IsFpSemigroup], 990, function(S)
       Print("<at least C(3) small overlap fp semigroup on the generators ",
             FreeGeneratorsOfFpSemigroup(S),">");
end);

InstallMethod(ViewObj, "for C2 finitely presented monoids",
              [IsC2 and IsFpMonoid], 980, function(S)
       Print("<at least C(2) small overlap fp monoid on the generators ",
             FreeGeneratorsOfFpMonoid(S),">");
end);

InstallMethod(ViewObj, "for C2 finitely presented semigroups",
              [IsC2 and IsFpSemigroup], 980, function(S)
       Print("<at least C(2) small overlap fp semigroup on the generators ",
             FreeGeneratorsOfFpSemigroup(S),">");
end);

InstallMethod(ViewObj, "for C1 finitely presented monoids",
              [IsC1 and IsFpMonoid], 970, function(S)
       Print("<at least C(1) small overlap fp monoid on the generators ",
             FreeGeneratorsOfFpMonoid(S),">");
end);

InstallMethod(ViewObj, "for C1 finitely presented semigroups",
              [IsC1 and IsFpSemigroup], 970, function(S)
       Print("<at least C(1) small overlap fp semigroup on the generators ",
             FreeGeneratorsOfFpSemigroup(S),">");
end);

#############################################################################
# A call to the function to PreferSmallOverlapEqualityTest() installs new
# methods for the equality test operator for elements of finitely presented
# monoids and semigroups respectively, which check if the containing monoid or
# semigroup satisfies C4, and apply C4WordProblemReckless() if applicable.
# This is not installed by default since checking the C4 condition imposes an
# extra overhead on equality tests in non-C4 semigroups

InstallGlobalFunction(PreferSmallOverlapEqualityTest, function(arg)
    local tester, verbose;
    if Length(arg) = 0 then
        verbose := false;
    elif Length(arg) = 1 then
        verbose := arg[1];
        if not IsBool(verbose) then
            Error("Argument to PreferSmallOverlapEqualityTest should be a boolean (or absent)");
        fi;
    else
        Error("PreferSmallOverlapEqualityTest takes zero or one arguments");
    fi;

    if verbose then 
        tester := function(u,v)
            local su;
            su := FpGrpMonSmgOfFpGrpMonSmgElement(u);
            if not IsC4(su) then
                Print("(using GAP's default (Knuth-Bendix?) equality test)\n");
                TryNextMethod();
            fi;
            Print("(using SmallOverlap equality test)\n");
            if su <> FpGrpMonSmgOfFpGrpMonSmgElement(v) then
                # In this case the arguments are not elements of the same
                # finitely presented semigroup or monoid, and so cannot be
                # equal.
                return false;
            fi;
            # Note that since we have already checked all the relevant
            # preconditions, it is okay to use C4WordProblemReckless(). 
            return C4WordProblemReckless(su, u, v);
        end;
    else 
        tester := function(u,v)
            local su;
            su := FpGrpMonSmgOfFpGrpMonSmgElement(u);
            if not IsC4(su) then
                TryNextMethod();
            fi;
            if su <> FpGrpMonSmgOfFpGrpMonSmgElement(v) then
                # In this case the arguments are not elements of the same
                # finitely presented semigroup or monoid, and so cannot be
                # equal.
                return false;
            fi;
            # Note that since we have already checked all the relevant
            # preconditions, it is okay to use C4WordProblemReckless(). 
            return C4WordProblemReckless(su, u, v);
        end;
    fi;

    InstallMethod(\=,
                  "for finitely presented semigroups satisfying C(4)",
                  [IsElementOfFpSemigroup, IsElementOfFpSemigroup],
                  tester
                 );

    InstallMethod(\=,
                  "for finitely presented monoids satisfying C(4)",
                  [IsElementOfFpMonoid, IsElementOfFpMonoid],
                  tester
                 );

    Print("SmallOverlap routines will now be used (where applicable) for\n");
    Print("testing equality in finitely presented semigroups and monoids.\n");
    if verbose then
        Print("And you'll know about it.\n");
    fi;
end);


##########################################################################
# C4WordProblem(u,v) decides (very efficiently) whether the elements
# u and v are equal in the C(4) finitely presented semigroup or monoid
# S. It returns fail if u and v are not elements of the same finitely
# presented monoid or semigroup, or if the containing semigroup does not
# satisfy C(4).

InstallGlobalFunction(C4WordProblem, function(uel,vel)
    local S;
    S := FpGrpMonSmgOfFpGrpMonSmgElement(uel);
    if not IsC4(S) then
        return fail;
    fi;
    if S <> FpGrpMonSmgOfFpGrpMonSmgElement(vel) then
        return fail;
    fi;
    return C4WordProblemReckless(S, uel, vel);
end);


##########################################################################
# C4WordProblemReckless(S,uel,vel) decides (very efficiently) whether the
# elements uel and vel are equal in the C(4) finitely presented semigroup
# or monoid S. S *MUST* be a finitely presented monoid or semigroup whose
# defining presentation satisfies C4, and uel and vel must be elements of
# it. Beware that no checks are made on these preconditions, and behaviour
# is unspecified if they are not satisfied (hence the name). This can be
# used in preference to C4WordProblem for efficiency, but should only be
# ventured when one knows with certainty that S, uel and vel satisfy the
# preconditions.

InstallGlobalFunction(C4WordProblemReckless, function(S,uel,vel)
    local relwords, u, v, p, upos, vpos, ppos,
          maxpieceprefixlengths, middlewordlengths, maxpiecesuffixlengths,
          maxpiecesuffixlengthbound, relwordlengths,
          ucop, vcop, ucopcomp,
          uhaswholeprefix, vhaswholeprefix,
          m;

    maxpieceprefixlengths := MaxPiecePrefixLengths(S);
    middlewordlengths := MiddleWordLengths(S);
    maxpiecesuffixlengths := MaxPieceSuffixLengths(S);
    maxpiecesuffixlengthbound := MaxPieceSuffixLengthBound(S);
    relwords := RelationWords(S);
    relwordlengths := RelationWordLengths(S);

    # We place list representations of u and v in a buffer with a little
    # extra space for working. TODO: Check this is big enough!
    u := Concatenation(List([1..maxpiecesuffixlengthbound], x -> 0),
                     LetterRepAssocWord(UnderlyingElement(uel)));
    v := Concatenation(List([1..maxpiecesuffixlengthbound], x -> 0),
                     LetterRepAssocWord(UnderlyingElement(vel)));
    upos := maxpiecesuffixlengthbound + 1;
    vpos := maxpiecesuffixlengthbound + 1;

    p := []; ppos := 1;

    repeat
        # If none of u or none of v is left, we need none of u, v
        # and p to be left for equality.
        if upos > Length(u) or vpos > Length(v) then
            return upos > Length(u) and vpos > Length(v) and ppos > Length(p);
        fi;

        ucop := CleanOverlapPrefix(S, u, upos);

        if ucop = fail then
            # If $u$ has no clean overlap prefix then for equality, u, v and
            # p (if non-empty) must all agree on the first letter.
            if u[upos] <> v[vpos] then
                return false;
            fi;
            if ppos <= Length(p) and u[upos] <> p[ppos] then
                return false;
            fi;
            # Assuming they do all agree on the first letter, then u = v
            # and p is a possible prefix if and only if the corresponding
            # statement holds for the suffixes of length n-1.
            upos := upos + 1;
            vpos := vpos + 1;
            ppos := ppos + 1; # Doesn't matter if this gets bigger than p
        else
            # We are now in the case that u has a clean overlap prefix,
            # in which case v must have either the same relation prefix,
            # or the one from the corresponding relation word
            ucopcomp := ComplementRelationWord(ucop);
            vcop := RelationPrefix(S, v, vpos);
            if vcop <> ucop and vcop <> ucopcomp then 
                return false;
            fi;

            # To be a possible prefix, p must be a prefix of either
            # the clean overlap prefix of u or that of its complement
            # relation word.
            if not IsFactorAt(relwords[ucop], 1, p, ppos)
                   and not IsFactorAt(relwords[ucopcomp], 1, p, ppos) then
                return false;
            fi;

            uhaswholeprefix := IsFactorAt(u, upos, relwords[ucop], 1);
            vhaswholeprefix := IsFactorAt(v, vpos, relwords[vcop], 1);

            if ucop = vcop and uhaswholeprefix and vhaswholeprefix then
                p := []; ppos := 1;
                if IsPrefixActive(S,
                                  u,
                                  upos+maxpieceprefixlengths[ucop]
                                             +middlewordlengths[ucop],
                                  upos+relwordlengths[ucop]) then
                    upos := upos + maxpieceprefixlengths[ucop]
                                                   + middlewordlengths[ucop];
                    vpos := vpos + maxpieceprefixlengths[ucop]
                                                   + middlewordlengths[ucop];
                else
                    upos := upos + relwordlengths[ucop]
                                            - maxpiecesuffixlengths[ucopcomp];
                    WriteMaxPieceSuffixInList(S, u, upos, ucopcomp);
                    vpos := vpos + relwordlengths[ucop]
                                            - maxpiecesuffixlengths[ucopcomp];
                    WriteMaxPieceSuffixInList(S, v, vpos, ucopcomp);
                fi;
            elif ucop = vcop then
                upos := upos + maxpieceprefixlengths[ucop]
                                                    + middlewordlengths[ucop];
                vpos := vpos + maxpieceprefixlengths[vcop]
                                                    + middlewordlengths[vcop];

                if IsFactorAt(relwords[ucop], 1, p, ppos) then
                    p := []; ppos := 1;
                else
                    # Set p to be the maximum piece suffix of relation
                    # word ucop
                    p := List([1..maxpiecesuffixlengths[ucop]],
                               x -> relwords[ucop][x
                                                   +maxpieceprefixlengths[ucop]
                                                   +middlewordlengths[ucop]]);
                    ppos := 1;
                fi;
           elif uhaswholeprefix and vhaswholeprefix then
                p := []; ppos := 1;
                if IsPrefixActive(S, u,
                                  upos+maxpieceprefixlengths[ucop]
                                                  +middlewordlengths[ucop],
                                  upos+relwordlengths[ucop]) then
                    upos := upos + maxpieceprefixlengths[ucop]
                                                 + middlewordlengths[ucop];
                    vpos := vpos + relwordlengths[vcop]
                                                 - maxpiecesuffixlengths[ucop];
                    WriteMaxPieceSuffixInList(S, v, vpos, ucop);
                else
                    vpos := vpos + maxpieceprefixlengths[vcop]
                                                 + middlewordlengths[vcop];
                    upos := upos + relwordlengths[ucop]
                                                 - maxpiecesuffixlengths[vcop];
                    WriteMaxPieceSuffixInList(S, u, upos, vcop);
                fi;
            elif vhaswholeprefix then
                p := []; ppos := 1;
                upos := upos + maxpieceprefixlengths[ucop]
                                                 + middlewordlengths[ucop];
                vpos := vpos + relwordlengths[vcop]
                                                 - maxpiecesuffixlengths[ucop];
                WriteMaxPieceSuffixInList(S, v, vpos, ucop);
            elif uhaswholeprefix then
                p := []; ppos := 1;
                vpos := vpos + maxpieceprefixlengths[vcop]
                                                + middlewordlengths[vcop];
                upos := upos + relwordlengths[ucop]
                                                - maxpiecesuffixlengths[vcop];
                WriteMaxPieceSuffixInList(S, u, upos, vcop);
            else
                # This is a very complicated case
                m := MaxCommonSuffixLengthOfRelationWords(S)[ucop][vcop];
                if not CompareFactors(u, upos, relwords[ucop], 1,
                                                  relwordlengths[ucop]-m) then
                     return false;
                fi;
                if not CompareFactors(v, vpos, relwords[vcop], 1,
                                                  relwordlengths[vcop]-m) then
                     return false;
                fi;
                upos := upos + relwordlengths[ucop] - m;
                vpos := vpos + relwordlengths[vcop] - m;
                # p becomes the maximal common suffix of relation words;
                p := List([1..m],
                          x -> relwords[ucop][relwordlengths[ucop] - m + x]);
                ppos := 1;
            fi;
        fi;
    until false;
end);

##########################################################################
# ComplementRelationWord(n) takes the number of a relation word and
# returns the number of the relation word on the other side of the
# presentation.

InstallGlobalFunction(ComplementRelationWord, function(n)
    if RemInt(n,2) = 0 then
        return n-1;
    else
        return n+1;
    fi;
end);

##########################################################################
# WriteMaxPieceSuffixInlist(S, u, upos, i) writes the max piece suffix of
# relation word i into the list u starting at position upos. TODO: Should
# probably be implemented differently

InstallGlobalFunction(WriteMaxPieceSuffixInList, function(S, u, upos, i)
   local relword, suffixlength, rellength, j;

   relword := RelationWords(S)[i];
   suffixlength := MaxPieceSuffixLengths(S)[i];
   rellength := RelationWordLengths(S)[i];
   for j in [1..suffixlength] do
       u[upos+j-1] := relword[rellength-suffixlength+j];
   od;
end);

##########################################################################
# IsPrefixActive(S, u, uposa, uposb) returns true if the suffix of list
# u beginning at position uposa contains a relation word prefix (see below
# or the paper for definition) beginning before (later) position uposb 

InstallGlobalFunction(IsPrefixActive, function(S, u, uposa, uposb);
    return FirstRelationPrefixPosition(S,u,uposa) < uposb;
end);

#####################################################################
# CleanOverlapPrefix(S, u, upos) returns the number of the relation word R
# such that the suffix of u starting at position upos begins with the
# max piece prefix of R followed by the middle word of R *AND* this factor
# s not overlapped by another relation prefix factor. it returns fail if
# there is no such relation word. It is only designed to work with
# C4 semigroups and monoids. Compare with the function RelationPrefix.

InstallGlobalFunction(CleanOverlapPrefix, function(S, u, upos)
    local maxpieceprefixlengths, middlewordlengths, i;
    maxpieceprefixlengths := MaxPiecePrefixLengths(S);
    middlewordlengths := MiddleWordLengths(S);
    i := RelationPrefix(S, u, upos);
    if i = fail then
        # There is no relation prefix at all, let alone a clean overlap one
        return fail;
    fi;
    if upos = Length(u) then
        # We have a relation prefix, and there is no space for an
        # overlapping relation prefix factor, so it must be a clean overlap
        # prefix
        return i;
    fi;
    if FirstRelationPrefixPosition(S, u, upos+1)
                  < upos+maxpieceprefixlengths[i] + middlewordlengths[i] then
        # We have a relation prefix, but it is overlapped by the next one
        # and so is not a clean overlap prefix
        return fail;
    fi;
    return i;
end);


########################################################################
# RelationPrefix(S, u, upos) returns the number of the relation word R
# such that the suffix of u starting at position upos begins with the
# max piece prefix of R followed by the middle word of R, or fail if
# there is no such relation word. It is only designed to work with
# C4 semigroups and monoids.

InstallGlobalFunction(RelationPrefix, function(S, u, upos)
    local relwords,
           maxpieceprefixlengths, middlewordlengths, maxpiecesuffixlengths,
           i, p, j;

    relwords := RelationWords(S);
    maxpieceprefixlengths := MaxPiecePrefixLengths(S);
    middlewordlengths := MiddleWordLengths(S);
    maxpiecesuffixlengths := MaxPieceSuffixLengths(S);

    for i in [1..Length(relwords)] do
        p := maxpieceprefixlengths[i] + middlewordlengths[i];

        # if u is long enough that this could be a prefix
        if p <= Length(u)-upos+1 then
            j := 1;

            # The following code should really be replaced with a
            # call to some general list testing function
            while j <= p and u[j+upos-1] = relwords[i][j] do
                j := j + 1;
            od;
            if j > p then
                # the relevant part of this relation word forms a prefix of
                # u starting at upos
                return i;
            fi;
        fi;
    od;
    # If we have get here then no relation prefix is a prefix of u
    # starting at position upos.
    return fail;
end);

########################################################################
# FirstRelationPrefixPosition(S, u, upos) returns the position of the
# start of the first factor of u which consists of the max piece prefix
# and middle word of a relation word, or fail if there is no such factor.
# This could doubtless be implemented much more efficiently

InstallGlobalFunction(FirstRelationPrefixPosition, function(S, u, upos)
   local i;
   for i in [upos..Length(u)] do
       if RelationPrefix(S, u, i) <> fail then
           return i;
       fi;
   od;
   return fail;
end);
